// *************************************************************************
//    Term.i
// *************************************************************************


#include "BindEnv.h"
#include "Debug.h"

// *************************************************************************
//
// A Term is implemented using a variant of structure sharing (though this
// may change):
// The 'expr' field contains the static structure of the term,
// while 'bindenv' maps (some of) the variables in 'expr' to terms.
// Variables in these `terms' are not interpreted in BindEnv, unlike what is
// done in Boyer and Moore structure sharing.
//


// *************************************************************************
//  Inline functions defined here.  They cannot be defined earlier since
//  they need to know the size of classes defined later.
// *************************************************************************
inline Cor_Term Cor_Term::dereference_var() const
{
    ASSERT(expr.is_var());

    return( bindenv->lookup(expr.var_num()) );
}

inline Cor_Term Cor_QTerm::dereference_var() const
{
    ASSERT(expr->is_var());

    return( bindenv->lookup(expr->var_num()) );
}

// *************************************************************************
//  The macro ARITH_DEREFERENCE_TERM overwrites its argument.
//  Use it carefully.  It takes as argument a term, and if it is an arithmetic
//  expression it is simplified.
// *************************************************************************

inline void Cor_arith_dereference_term(Cor_Term &term)
{
    Cor_TermLink *renamed_vars = NULL;
    Cor_Term newt;
    term.expr.simplify(newt.expr, term.bindenv, renamed_vars,
               term.bindenv, NULL);
    if (! newt.expr.is_empty() ) {
      newt.bindenv = term.bindenv;
      term = newt;
    }
}

// *************************************************************************
//  The function full_dereference_term overwrites its argument.
//  Use it carefully.
//  It takes as argument a term, and dereferences the term fully, till the
//  term is either a free variable or a non-variable structure or a constant.
//  Note that this merely chases down chains of variable bindings. It does NOT
//  go into structures.
// *************************************************************************
inline void Cor_full_dereference_term(Cor_Term &term)
{
    if (!term.expr.is_empty()) {

        while (term.bindenv && term.expr.is_var()) {

            if (! term.bindenv->try_lookup(term,term.expr.var_num()))
                break;
        }

        if (term.expr.is_expr())
            Cor_arith_dereference_term(term);
    }
}

// *************************************************************************
// The function deep_dereference_term overwrites its argument.
// Use it carefully.
// It takes as argument a term, and deeply dereferences the term, by replacing
// all bound variables by the appropriate constant.
// *************************************************************************
inline void Cor_deep_dereference_term(Cor_Term &term)
{
    Cor_full_dereference_term(term);

    if (!term.expr.isConstant() && term.bindenv) {
        
        if (term.expr.simplify_me(term.bindenv))
          term.bindenv = NULL;
    }
}

// *************************************************************************
//  isfree_term - returns 1 if argument is a free variable, 0 otherwise
// *************************************************************************
inline  int Cor_isfree_term(Cor_Term *t)
{
  return (t->expr.is_var() &&
            t->bindenv->not_bound(t->expr.var_num()));
}

// *************************************************************************
