// *************************************************************************
//    Compile.C
// *************************************************************************

#include <stdio.h>
#include "Compile.h"

#include "CompileRule.h"
#include "Unparse.h"
#include "ParserStruct.h"
#include "Execute.h"
#include "Globals.h"
#include "Literal.h"
#include "SccInfo.h"
#include "BuiltinSymbols.h"
#include "ExportInfo.h"
#include "DerivedMethod.h"
#include "QFPipelinedModuleData.h"
#include "Pipelined.h"
#include "Database.h"
#include "Collection.h"
#include "Debug.h"

#ifdef ASYN_SOCK_REL
#include "RelationMisc.h"
#endif

// *************************************************************************

Cor_Scc *Cor_SccList = NULL;
Cor_Scc **Cor_SccListTail = &Cor_SccList;

Cor_SccWork *Cor_SccWorkList = NULL;
int Cor_SccCount = 0;
Cor_SccWork Cor_SccDone[1]; // dummy

// *************************************************************************
int Cor_get_offset(Cor_Table *F, Cor_CArg& name, int arity, int print_error)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;

    for (i = 0; i< F->numRels; i++) {
        if ((F->rels[i].relArity == arity)
                && (F->rels[i].relName.equals(name)))
            break;
    }
    
    // The offset returned should be within the array bounds.
    if (i >= F->numRels) {
        if (print_error) {
            Cor_exEnv->err_strm->printf(
                "Error: Couldn't find relation %s with arity %d\n",
                name.symbol(), arity);
        }
        return(-1);
    }
    ASSERT((i < F->numRels));
    return(i);
}

// *************************************************************************
//    NOTE: Predicate field in Literal is often not set !! So use pred field.

int Cor_get_offset(Cor_Table *F, Cor_Literal *pred, int print_error)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return Cor_get_offset(F, pred->pred, pred->arity(), print_error);
}

// *************************************************************************
//  This function finds the offset of an RMark in the RMarkArray for an scc
//  It depends on the order in which the RMarks are actually allocated, which
//  happens in SccInfo::Solve(). The function returns the offset of the first
//  RMark for that predicate (i.e.) the old RMark. The new RMark is in the
//  adjacent position.
// *************************************************************************

int Cor_Find_RMark_Offset(Cor_SccInfo *scc_info, Cor_Predicate *rpred)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Predicate **pred_ptrs = scc_info->pred_ptrs;
#ifdef ASYN_SOCK_REL
    int total_numpreds = scc_info->numPreds + scc_info->num_lowerPreds + scc_info->numSockRelPreds;
#else
    int total_numpreds = scc_info->numPreds + scc_info->num_lowerPreds;
#endif
    
    for (int i=0; i < total_numpreds; i++) {
        if (pred_ptrs[i] == rpred)
            return(2*i);
    }

    // Default --- this should never be reached !!
    Cor_exEnv->err_strm->printf(
        "CORAL:: Error - Find_RMark_Offset - can't find %s(arity %d)\n",
        rpred->name.symbol(), rpred->arity);
    return(-1);
}

// *************************************************************************
int Cor_has_fail_in_body(Cor_Clause *rule)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;

    for (i=0; i < rule->num_preds; i++) {
        if (rule->preds[i]->pred.equals(Cor_FailSymbol))
            return(i);
    }
    return(-1);
}

// *************************************************************************
void Cor_add_pred_extern(Cor_Predicate *pred, enum PredKind kind,
    Cor_ModulePred *predicates)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    Cor_CArg name = pred->name;
    int arity = pred->arity;
    Cor_ModulePred *new_pred;

    for (new_pred = predicates->next; new_pred; new_pred = new_pred->next) {
        if (    (new_pred->pred->arity == arity)
             &&  new_pred->pred->name.equals(name)) {

            if ( (kind == LocalPredKind)
                  && (new_pred->kind == ExternalPredKind))
                new_pred->kind = kind;
            return;
        }
    }

    new_pred = new Cor_ModulePred(pred);
    new_pred->kind = kind;
    new_pred->next = predicates->next;
    predicates->next = new_pred;
}

// *************************************************************************
static Cor_Scc *NewScc()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Scc *scc = new Cor_Scc();
    scc->preds = NULL;
    scc->next = NULL;
    scc->recursive = 0;
    *Cor_SccListTail = scc;
    Cor_SccListTail = &scc->next;
    scc->emitted = -1;
    return(scc);
}

// *************************************************************************
void Cor_initialize_scc_structs()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    Cor_SccList = NULL;
    Cor_SccListTail = &Cor_SccList;

    Cor_SccWork *curr, *prev = NULL;

    for (curr = Cor_SccWorkList; curr; curr = curr->next) {
        delete prev;
        prev = curr;
    }
    delete prev;

    Cor_SccWorkList = NULL;
    Cor_SccCount = 0;
}

// *************************************************************************
void Cor_clear_scc_list()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

/*
      // Messes up print_module.
    Cor_Scc *curr, *prev = NULL;
    for (curr = Cor_SccList; curr != NULL; curr = curr->next) {
        delete prev;
        prev = curr;
    }
    delete prev;
*/

    Cor_SccList = NULL;
    Cor_SccListTail = &Cor_SccList;
}

// *************************************************************************
inline int min(int i, int j)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return i < j ? i : j;
}

// *************************************************************************
int Cor_find_predicates_invoked_from(Cor_Predicate *pred, int original)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Based on code by Thomas Ball from an algorithm by Robert Tarjan

    Cor_SccWork *work = new Cor_SccWork();
    Cor_Clause  *clause;
    int ipred;
    // number is a sequential node number
    // lowlink is lowest number of a node reachable from current node.
    work->lowlink = work->number = ++Cor_SccCount;

    work->next = Cor_SccWorkList;
    work->pred = pred;
    work->recursive = 0;
    pred->work = work;
    Cor_SccWorkList = work;

#ifdef ASYN_SOCK_REL
    Cor_Scc *scc = NewScc();
#endif
    
    // Iterate over all Clauses that define 'pred'.
    for (clause = pred->clauses; clause; clause = clause->pnext) {
        // This should take care of 'bogus' magic dependencies
        if (clause->head->predicate->Cor_is_magic() && clause->num_preds) {
            Cor_Predicate *first_pred = clause->preds[0]->predicate;
            if (first_pred->Cor_is_magic() && first_pred->mark == 0)
                continue;
        }

      // when we are constructing the original Scc structure,
      // we ignore rewritten rules, i.e. rules with 2 labels
      if (original && (clause->clause_names->count() >= 2))
        continue;

        clause->seen = 1;
        // Iterate over all Predicates invoked by 'clause'.
        for (ipred = 0; ipred < clause->num_preds; ipred++) {
            Cor_Literal *child_literal = clause->preds[ipred];
            Cor_Predicate *child = child_literal->predicate;
            if (child->work == NULL) { /*  not visited yet */
                int child_lowlink = Cor_find_predicates_invoked_from(child);
                work->lowlink = min(work->lowlink, child_lowlink);
            }
            else if (child->work != Cor_SccDone) {
                if (child->work->number < work->number)
                    work->lowlink = min(work->lowlink, child->work->lowlink);
                child->work->recursive = 1;
            }
        }

#ifdef ASYN_SOCK_REL
        // for asyn_socket_rel
        for (ipred = 0; ipred < clause->num_preds; ipred++) {
            Cor_Literal *child_literal = clause->preds[ipred];
            Cor_Predicate *child = child_literal->predicate;
            if (child->asyn_sock()) {
                scc->recursive = 1;
            }
        }
#endif
    }

    int lowlink = work->lowlink;
    if (work->lowlink == work->number) {
#ifndef ASYN_SOCK_REL
        Cor_Scc *scc = NewScc();
#endif
        Cor_Predicate **pred_tail = &(scc->preds);
        int work_num = work->number;
        while (Cor_SccWorkList
                    && (Cor_SccWorkList->number >= work_num)) {
            Cor_SccWork *sccWork = Cor_SccWorkList;
            Cor_Predicate *work_pred = sccWork->pred;
            work_pred->next = scc->preds;
            work_pred->scc = scc;
#            ifdef DEBUG
                if (work_pred->work != sccWork)
                    ASSERT(0);
#            endif
            work_pred->work = Cor_SccDone;
            *pred_tail = work_pred;
            pred_tail = &work_pred->next;
            Cor_SccWorkList = sccWork->next;
            scc->recursive |= work->recursive;
            delete sccWork;
        }

        *pred_tail = NULL;
    }
    return lowlink;
}

// *************************************************************************
static void mark_predicates_invoked_from(Cor_Predicate *pred)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Clause *clause;
    int ipred;

    pred->mark = 1;
    // Iterate over all Clauses that define 'pred'.
    for (clause = pred->clauses; clause; clause = clause->pnext) {
        if (clause->head->predicate->Cor_is_magic()) continue;

        // Iterate over all Predicates invoked by 'clause'.
        for (ipred = 0; ipred < clause->num_preds; ipred++) {
            Cor_Literal *child_literal = clause->preds[ipred];
            Cor_Predicate *child = child_literal->predicate;
            if (child->mark == 0) { // not visited yet
                mark_predicates_invoked_from(child);
            }
        }
    }
}

// *************************************************************************
static void clean_up_after_one_scc()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

      // if there are no base predicates, need to clean up SccList
    Cor_Scc *base_scc = Cor_SccList;
    if (!(base_scc->preds)) {
        // there are no base predicates
        ASSERT((base_scc->next != NULL));

        // must at least have the query predicate
        ASSERT((base_scc->next->preds != NULL));

        Cor_SccList = base_scc->next;
        delete base_scc;
    }
}

// *************************************************************************
int Cor_group_predicates_in_one_scc(Cor_Predicate *pred)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // If there are base predicates, there have to be at least two
    // SCCs; the first one must be the base SCC

    Cor_Scc *base_scc;
    Cor_Scc *scc;

    if (!Cor_SccList) {
        base_scc = NewScc();
        scc = NewScc();         // this should link it to be the 2nd SCC
        scc->recursive = 1;        // NOTE: is this a right assumption?
    } else {
        base_scc = Cor_SccList;
        scc = Cor_SccList->next;
    }

    Cor_SccWork *work;
    if (!Cor_SccWorkList) {
        work = new Cor_SccWork();
        work->next = Cor_SccWorkList;
        Cor_SccWorkList = work;
    } else
        work = Cor_SccWorkList;

    pred->work = work;  // visited

    Cor_Clause *clause;
    int ipred;
    
    pred->scc = scc;
    pred->next = scc->preds;
    scc->preds = pred;

    for (clause = pred->clauses; clause; clause = clause->pnext) {
        // This should take care of 'bogus' magic dependencies
        if (clause->head->predicate->Cor_is_magic() && clause->num_preds) {
            Cor_Predicate *first_pred = clause->preds[0]->predicate;
            if (first_pred->Cor_is_magic() && first_pred->mark == 0)
                continue;
        }

        clause->seen = 1;

        // Iterate over all Predicates invoked by 'clause'.
        for (ipred = 0; ipred < clause->num_preds; ipred++) {
            Cor_Literal *child_literal = clause->preds[ipred];
            Cor_Predicate *child = child_literal->predicate;
            if ((child->work == NULL) && (child->clauses)) {
                // predicates not yet visited and those
                // which have rules defined in this module
                Cor_group_predicates_in_one_scc(child);
            }
            else if (child->work == NULL) {
                // predicates not yet visited
                // and which are external predicates
                child->work = work;
                if (child->Cor_is_mp_done()) { 
                    // NOTE: for Ordered Search
                    child->scc = scc;
                    child->next = scc->preds;
                    scc->preds = child;
                }
                else {
                    child->scc = base_scc;
                    child->next = base_scc->preds;
                    base_scc->preds = child;
                }
            }
        }
    }

    return 1;
}

// *************************************************************************
static void interpret_scc(Cor_QFMaterializedModuleData *md, Cor_Scc *scc,
    Cor_ExportInfo *export_info, Cor_Predicate *ext_magicpred)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Predicate *pred; Cor_Clause *clause;
    int numRules = 0;

    // First emit code to solve Scc's needed by this Scc.
    // (This is, after all, bottom-up evaluation.)

    for (pred = scc->preds; pred != NULL; pred = pred->next) {
        for (clause = pred->clauses;
            clause != *pred->clause_tail;
            clause = clause->pnext) {
            if (!clause->seen) continue;
            int ipred;
            numRules++;
            for (ipred = 1; ipred <= clause->num_preds; ipred++) {
                Cor_Predicate *child = clause->preds[ipred-1]->predicate;
                if (child->scc == scc) continue;
                if (child->scc->emitted == export_info->sequence)
                    continue; // Already dumped.
                child->scc->emitted = export_info->sequence;
                interpret_scc(md, child->scc, export_info, ext_magicpred);
            }
        }
    }

    // Check that the scc has some rules to evaluate. Otherwise, it is useless
    if (numRules)  md->AddScc(scc, ext_magicpred);
}

// *************************************************************************
Cor_Predicate *Cor_find_ext_magicquerypred(Cor_ParserStruct *Cor_parserStruct)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    for (Cor_Predicate *pred = Cor_parserStruct->AllPredicates; pred != NULL;
        pred=pred->allprednext)
    {
        if ((!pred->clauses) && (pred->Cor_is_magic())) { // found it
            return pred;
        }
    }
    return NULL;
}

// *************************************************************************
int Cor_determine_subsum_status(Cor_QFModuleData *md, Cor_CArg& name,
        int default_val)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_PredAnnotations *cur;
    
    for (cur= md->multiset_annotations; cur; cur=cur->next()) {
        char buf[1000];
        Cor_sprint_pred_name(buf, cur->pred.symbol(), cur->adorn);
        // Warning: multiset and checksubsumption annotations have
        // the name stored separate from the adornment 
        if (name.equals(buf)) {
            default_val = cur->check_subsum;
            break;
        }
    } 

    /***
    For pipelined modules, this used to be the code : Praveen

    for (cur= md->multiset_annotations; cur; cur=cur->next()) {
        if (cur->pred==rule->head->pred)
             F->rels[offset].relPtr->check_subsum = cur->check_subsum;
    }
    ****/

    return default_val;
}

// *************************************************************************
static int diskrel_anno(Cor_QFModuleData *md, Cor_CArg& name,
        Cor_ArgList *&a_list)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_PredAnnotations *cur;

    // what about the schema arglist in this case ?? : PRAVEEN
    if (md->module_info->DiskRel)
        return 1;
    
    for (cur= md->diskrel_annotations; cur; cur=cur->next()) {
        char buf[1000];
        Cor_sprint_pred_name(buf, cur->pred.symbol(), cur->adorn);
        // Warning: diskrel annotations have the name stored separate 
        // from the adornment 
        if (name.equals(buf)) {
            a_list = cur->arglist1;
            return 1;
        }
    }

    return 0;
}

// *************************************************************************
void Cor_get_annos(Cor_QFModuleData *md, Cor_ModulePred *predicate)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_PredAnnotations *cur;
    Cor_AnnoStruct *new_anno;

    for (cur= md->index_annotations; cur; cur = cur->next()) {
        if (cur->pred.equals(predicate->pred->name)) {

            int mx1, mx2 = 0;
            mx1 = cur->arglist1->max_vars();
            if (cur->arglist2 )
                mx2 = cur->arglist2->max_vars();
              
            int num_vars = (mx1 > mx2) ? mx1 : mx2;
            new_anno = new Cor_AnnoStruct(cur, num_vars,
                predicate->index_annos);
            predicate->index_annos = new_anno;
        }
    }

    for (cur= md->agg_sel_annotations; cur; cur=cur->next()) {
        if (cur->pred.equals(predicate->pred->name)) {
           new_anno = new Cor_AnnoStruct(cur,0,predicate->aggsel_annos);
           predicate->aggsel_annos = new_anno;
        }
    }
}

// *************************************************************************
static void init_module_preds(Cor_QFModuleData *md)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_ModulePred *predicates = md->predicates->next;

    for (; predicates; predicates = predicates->next) {
        Cor_CArg name = predicates->pred->name;
        int arity = predicates->pred->arity;
        predicates->offset = Cor_get_offset(md->F, name, arity, 1);

        switch (predicates->kind) {
        case QueryPredKind:
            predicates->disk_rel =
                diskrel_anno(md, name, predicates->schema_def);
            predicates->check_subsum =
                Cor_determine_subsum_status(md, name,
                    md->module_info->CheckSubsum);
            break;

        case LocalPredKind:
            predicates->disk_rel =
                diskrel_anno(md, name, predicates->schema_def);
            if (Cor_is_magic(name)) {
                 predicates->check_subsum = Cor_determine_subsum_status(md,name,
                    (md->module_info->CheckSubsum | md->module_info->MultiSet));
            }
            else {
                 predicates->check_subsum = Cor_determine_subsum_status(md,
                    name, md->module_info->CheckSubsum);
            }
            break;

        case MagicQueryPredKind:
            predicates->disk_rel = diskrel_anno(md, name,
                predicates->schema_def);
            predicates->check_subsum = Cor_determine_subsum_status(md, name,
                (md->module_info->CheckSubsum | md->module_info->MultiSet));
            break;

        case ExternalPredKind:
            break;
        }
        Cor_get_annos(md, predicates);
    }
}

// *************************************************************************
/*
    This routine creates the structures within the QFMaterializedModuleData,
    and inserts it into the IterativeMethod table. It assumes that SCC
    analysis has already been performed w.r.t. the particular exported
    predicate.
*/

static void interpret_materialized_qf(Cor_QFMaterializedModuleData *md,
    Cor_ExportInfo *export_info, Cor_ParserStruct *Cor_parserStruct)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    Cor_CArg exported_name = export_info->pred_name;

    Cor_Scc *scc;
    Cor_ModulePred *predicates = md->predicates;

    // Make sure that all Relations used by the query are initialized
    // by init_rule_pointers. Use add_pred_extern to enter them into
    // the predicates collection to keep track of them.
    // First enter the the query and magic predicates:

    Cor_Predicate *query_predicate = export_info->predicate;
    Cor_add_pred_extern(query_predicate, QueryPredKind, predicates);

    int query_arity = query_predicate->arity;
    int magic_arity = export_info->adornment->count();
    char *magic_name = new char[exported_name.symlen()+4+query_arity];
    Cor_sprint_magic_name(magic_name, exported_name.symbol(),
         export_info->adornment);

    Cor_CArg mname = Cor_CArg(magic_name);
    delete [] magic_name;
    Cor_Predicate *magic_predicate =
            Cor_LookupPredicate(Cor_parserStruct, mname, magic_arity);

    if (magic_predicate) {
        Cor_add_pred_extern(magic_predicate, MagicQueryPredKind, predicates);
    } else {
        *(Cor_exEnv->err_strm)
            << "Warning: No magic predicate for exported predicate "
            << exported_name.symbol() << endl;
    }

    // Then enter the remaining predicates.
    for (scc = Cor_SccList; scc != NULL; scc = scc->next) {
        Cor_Predicate *pred;
        Cor_Clause *clause;
        int ipred;
        Cor_Literal **literals;

        // Add predicates for non-recursive clauses in this scc.
        for (pred = scc->preds; pred != NULL; pred = pred->next) {
            clause = pred->clauses;
            // If pred->clauses == *pred->clause_tail, then
            // the pred has no clauses, and hence is an ExternalPredKind.
            if (clause != *pred->clause_tail) {
                Cor_add_pred_extern(pred, LocalPredKind, predicates);
            }
            else if (pred->Cor_is_mp_done()) {
                // NOTE: this is for ordered search
                Cor_add_pred_extern(pred, LocalPredKind, predicates);
            } else
                Cor_add_pred_extern(pred, ExternalPredKind, predicates);

            // This also determines if clause is recursive.
            for (; clause != *pred->clause_tail; clause = clause->pnext) {
                if (!clause->seen)
                    continue;
                clause->recursive = 0;

                for (literals = clause->preds, ipred = clause->num_preds;
                     --ipred>= 0; literals++)
                {
                    Cor_add_pred_extern(literals[0]->predicate,
                            ExternalPredKind, predicates);

                    // with Ordered Search, have to special case the 
                    // negated literals
                    if (Cor_parserStruct->CurModule->UseOrdSearch) {
                        if (literals[0]->negated)
                            continue;
                    }
                    // don't need to special case for done_m_p since 
                    // it is covered by the next if condition

                    if (literals[0]->predicate->scc == scc) 
                        clause->recursive = 1;
                }

#ifdef ASYN_SOCK_REL
                // for asyn_socket_rel
                literals = clause->preds;

                for (int j = 0; j < clause->num_preds; j++) {
                    if (literals[j]->predicate->asyn_sock()) {
                        clause->recursive = 1;
                    }
                }
#endif
            }
        }
    }

    md->export_pred = export_info->predicate;
    md->module_info = Cor_parserStruct->CurModule;
    Cor_parserStruct->CurModule = new Cor_ModuleInfo();

    interpret_scc(md, export_info->predicate->scc, export_info,
        Cor_find_ext_magicquerypred(Cor_parserStruct));

    // move some of the run time work from init_rule_pointers to
    // compile time here. Determine what flags need to be set, etc.
    init_module_preds(md);
    
    // Insert the module structure along with its associated exported
    // query form into the IterativeMethod table.
    if (! exported_name.is_empty()) {

        Cor_DerivedMethod *query_method;
       
        if (export_info->adornment) 
            query_method = new Cor_DerivedMethod(exported_name,
                    *(export_info->adornment), Cor_do_query,
                    (Cor_QFModuleData *)md); 
        else {
            Cor_BitVector bv;
            query_method = new Cor_DerivedMethod(exported_name,
                    bv, Cor_do_query, (Cor_QFModuleData *)md); 
        }

        if (query_method)    // To eliminate "unused" warning
            ;    // Nothing
    }

       // insert the module to ModuleTable in CurDB
   if (Cor_CurDB->ModuleTable == NULL)
       Cor_CurDB->ModuleTable = new Cor_SymTable();

   Cor_CurDB->ModuleTable->insert(md->module_info->name, md);
}

// *************************************************************************
static void create_module_table(Cor_ParserStruct *Cor_parserStruct,
    Cor_Table *&module_table, Cor_RuleInfoArray *&module_rules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    // Create the Table structure (containing a pointer to each Relation used).
    int pred_count = 0;
    Cor_Predicate *pred;

    for (pred = Cor_parserStruct->AllPredicates; pred != NULL;
        pred = pred->allprednext) {
        pred_count++;
    }

    module_table = new Cor_Table(pred_count);
    module_rules = new Cor_RuleInfoArray;

    // This enters the name and arity of each relation in the Table into the
    // Table array 'module_table'. The relations are also initialized at the
    // same time.

    int i = 0;
    for (pred = Cor_parserStruct->AllPredicates; pred != NULL;
            pred=pred->allprednext, i++) {
        module_table->AddValue(i, pred->name, pred->arity);
        pred->frame_offset = i;
    }

    // For each clause, create a rule-info structure that contains the 
    // information necessary to evaluate it.

    int rule_number;

    Cor_Clause *clause;
    int num_rules = Cor_parserStruct->cur_collect->count();

    for (rule_number = 0; rule_number < num_rules; rule_number++ ) {
        clause = Cor_parserStruct->cur_collect->get_ith(rule_number);
        
        clause->rule_number = rule_number;
        module_rules->AddRule(Cor_create_rule_info(module_table, clause));
    }
}

// *************************************************************************
//
//  Add sup_magic predicate to appropriate Scc. Used in NestedSccEval only
//
void Cor_add_sup_to_scc(Cor_ParserStruct *Cor_parserStruct)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
  // traverse the list of predicate, for each sup, find the scc it belongs to
  for (Cor_Predicate *pred = Cor_parserStruct->AllPredicates; pred != NULL;
        pred=pred->allprednext) 
    if (pred->Cor_is_supp()) {
      // get the first predicate in the rhs of clause, e.g. m_p
      Cor_Predicate *first_pred_in_clause = pred->clauses->preds[0]->predicate;
      // get the name of predicate which is in the same Scc as this sup
      // i.e. p
      char *name_of_pred_in_same_scc =
            first_pred_in_clause->name.symbol() + 2;

      Cor_CArg sccname = Cor_CArg(name_of_pred_in_same_scc);
      Cor_Predicate *pred_in_same_scc =
                Cor_LookupPredicate(Cor_parserStruct, sccname, -1);
      pred->scc = pred_in_same_scc->scc;
    }
}

// *************************************************************************
//
// Make a list of external predicates for each Scc.
// Used in NestedSccEval only.
//
void Cor_make_ext_pred_list(Cor_ParserStruct *Cor_parserStruct)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
  // For rules with body predicates that are derived and not in the 
  // same SCC as the head predicate, add those body predicates to the 
  // list of external predicates of the SCC of the head.

  Cor_Clause *rule;
  int j, nrules;

  nrules = Cor_parserStruct->cur_collect->count();

  for (j=0; j < nrules; j++) {
      rule = Cor_parserStruct->cur_collect->get_ith(j); 

        // check each predicate in the body
        for (int i=0; i < rule->num_preds; ++i) 
          if (!rule->preds[i]->base_pred &&
                  (rule->preds[i]->predicate->scc != 
                      rule->head->predicate->scc)) {
              // derived predicate in lower Scc
              // first check if it is already in the list
            int exist = 0;
            for (struct PredList *ext_preds =
                    rule->head->predicate->scc->ext_preds;
                    ext_preds != NULL;
                    ext_preds = ext_preds->next)
                if (ext_preds->pred == rule->preds[i]->predicate) {
                    exist = 1;
                    break;
                }
          }
    }
}

// *************************************************************************
//
// Top-level routine to interpret a CORAL module. For each exported query
// form, a QFModuleData structure is created that encapsulates the information
// needed to evaluate that query. The exported query form and the associated 
// QFModuleData structure are stored in a table of IterativeMethods.
//

void Cor_interpret_module(Cor_ParserStruct *Cor_parserStruct)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    Cor_ExportInfo *export_info = NULL;
    Cor_Table *module_table = NULL;
    Cor_RuleInfoArray *module_rules = NULL;

    if (Cor_parserStruct->CurModule->ExportedPreds == NULL) {
        *(Cor_exEnv->err_strm) << "CORAL: Error - No exported query form.\n";
        return;
    }

    // Create the Table structure (containing a pointer to each Relation used).
    create_module_table(Cor_parserStruct, module_table, module_rules);

    Cor_QFModuleData *module_data = NULL;

    // Process each exported query form independently.
    for (export_info = Cor_parserStruct->CurModule->ExportedPreds;
                export_info != NULL; export_info = export_info->next) {
        Cor_CArg query_symbol;
        Cor_CArg exported_name = export_info->pred_name;
        int query_arity;

        if (Cor_parserStruct->CurModule->UsePipelining) {
            module_data = new Cor_QFPipelinedModuleData(module_table,
                module_rules, Cor_parserStruct);
        } else {
            module_data = new Cor_QFMaterializedModuleData(module_table,
                module_rules, Cor_parserStruct);
        }

        if (export_info->adorn_symbol.is_empty()) {
            query_arity = 0;
            query_symbol = exported_name;
        } else {
            Cor_CArg exported_adorn_name = export_info->adorn_symbol;
            query_arity = (int) exported_adorn_name.symlen();
                
            if (Cor_parserStruct->CurModule->UsePipelining) {
                int len = (int) exported_name.symlen() +1;
                char *query_name = new char[len+1];
                sprintf(query_name,"%s",exported_name.symbol());
                query_symbol = Cor_CArg(query_name);
                delete [] query_name;
            } else {
                int len = (int) exported_name.symlen() +1 + query_arity;
                char *query_name = new char[len+1];
                   Cor_sprint_pred_name(query_name,
                        exported_name.symbol(),
                    exported_adorn_name.symbol());
                query_symbol = Cor_CArg(query_name);
                delete [] query_name;
            }
        }

        // Ensure that the exported query form is a valid one.
        export_info->predicate = Cor_LookupPredicate(Cor_parserStruct,
            query_symbol, query_arity);
        if (!export_info->predicate) {
            Cor_exEnv->err_strm->printf(
                "Error: Exported predicate %s not defined in module\n",
                query_symbol.symbol() );
            continue;
        }

        if (Cor_parserStruct->CurModule->UsePipelining) {
            Cor_interpretPipelined((Cor_QFPipelinedModuleData *) module_data,
                export_info, Cor_parserStruct);
        } else {
            // For MATERIALIZED modules
            // Do Scc (Strongly Connected Components) analysis.

            Cor_Clause *clause;
            int j, nrules;
            nrules = Cor_parserStruct->cur_collect->count();

            for (j=0; j < nrules; j++) {
                clause = Cor_parserStruct->cur_collect->get_ith(j);
                clause->seen = 0;
                clause->recursive = 0;
                for (int ipred = 0; ipred < clause->num_preds; ipred++) {
                    clause->preds[ipred]->predicate->work = NULL;
                    clause->preds[ipred]->predicate->mark = 0;
                }
            }
    
            Cor_initialize_scc_structs();

            if (Cor_parserStruct->CurModule->SingleScc
                    || Cor_parserStruct->CurModule->UseOrdSearch) {
                mark_predicates_invoked_from(export_info->predicate);
                Cor_group_predicates_in_one_scc(export_info->predicate);
                clean_up_after_one_scc();
            } else {
                mark_predicates_invoked_from(export_info->predicate);
                if (Cor_parserStruct->CurModule->NestedSccEval) {
                    Cor_find_predicates_invoked_from(export_info->predicate,1);
                    Cor_add_sup_to_scc(Cor_parserStruct);
                    Cor_make_ext_pred_list(Cor_parserStruct);
                } else
                    Cor_find_predicates_invoked_from(export_info->predicate);
            }
            interpret_materialized_qf(
                    (Cor_QFMaterializedModuleData *)module_data,
                    export_info, Cor_parserStruct);
        }
    }

    Cor_clear_scc_list();
}

// *************************************************************************
void Cor_compile_single_rule(Cor_ParserStruct *Cor_parserStruct,
        Cor_Table* &module_table, Cor_RuleInfoArray* &module_rules,
        Cor_SemiNaive &snrule)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    // Create the Table structure (containing a pointer to each Relation used).
    create_module_table(Cor_parserStruct, module_table, module_rules);

    // There is only one RuleInfo structure
    ASSERT((module_rules->numRules == 1));

    // There was only one rule in the parserStruct
    // SPF - This will change, with a module per parserStruct in
    // imperative mode?
    ASSERT((Cor_parserStruct->cur_collect->count() == 1));

    Cor_Clause *clause = Cor_parserStruct->cur_collect->get_ith(0);

    Cor_compile_nonrec_rule(snrule, clause, module_rules->rInfoPtrs[0]);
}

// *************************************************************************
void Cor_number_sccs(Cor_Scc *Sccs)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
  // number the SCCs after the scc analysis is done
  for (int scc_no = 0; Sccs != NULL; scc_no++, Sccs = Sccs->next)
    Sccs->scc_no = scc_no;
}

// *************************************************************************
