// *************************************************************************
//    StrRefTable.C  - Reference Counting for strings
// *************************************************************************

#include <string.h>

#include "StrRefTable.h"

#include "Globals.h"
#include "Debug.h"

// *************************************************************************
Cor_StrRefElement::Cor_StrRefElement()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    _str     = NULL;
    _ref_cnt = 0;
    _next    = NULL; 
}

// *************************************************************************
Cor_StrRefElement::Cor_StrRefElement(char *str, int init_cnt)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    _str = str;
    _ref_cnt = init_cnt;
    _hash = Cor_StrHash(str);

    _next = NULL;
}

// *************************************************************************
Cor_StrRefElement::~Cor_StrRefElement()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    delete [] _str;
}

// *************************************************************************
void Cor_StrRefElement::complain()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

#ifdef    DEBUG

    if (complain) {
        // Should complain here, and report the string and its current count?
        cerr << _str << ": has a ref count of " << _ref_cnt << endl;
    }

#endif

}

// *************************************************************************
Cor_StrRefCursor::Cor_StrRefCursor(Cor_StrRefTable *_table)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    table = _table;

        // The next attempt will get the first item from index 0
    index = 0;
    current = table->_reftab[0];

    if (! current)
        ScanNext();
}

// *************************************************************************
Cor_StrRefCursor::~Cor_StrRefCursor()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
}

// *************************************************************************
int Cor_StrRefCursor::ScanNext()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (index < 0)
        return(0);

    if (current){

        current = current->_next;
        if (current)
            return(1);
    }

    int limit = table->size();
    for (index++; index < limit; index++) {
        if ((current = table->_reftab[index]))
            return(1);
    }

    index = -1;
    return(0);
}

// *************************************************************************
Cor_StrRefTable::Cor_StrRefTable(int size)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (size == 0)
        size = COR_DEFLT_STRREF_TAB_SIZE;

    _size = size;
    _num_elements = 0;

        // Rehash larger when more than 2 elements (on average)
    _element_limit = _size + _size;

    _reftab = new Cor_StrRefElement*[_size];
    ASSERT(_reftab != NULL);

    for (int j=0; j < _size; j++)
        _reftab[j] = NULL;
}

// *************************************************************************
Cor_StrRefTable::~Cor_StrRefTable()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef    DBG_REF_CNT
    // Complain about things remaining in the table
    clear(1);
#else
    clear();
#endif

    delete [] _reftab;
}

// *************************************************************************
char *Cor_StrRefTable::insert(char *str, int **refer_pntr,
        int alloc_on_insert, HashVal *hash, int hash_valid)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(this != NULL);

    if (!this)
        return(str);

      // What about this, should we just return it?
    ASSERT(str != NULL);

    int indx;

    Cor_StrRefElement *element = lookup(str, &indx, hash, hash_valid);

    if (element) {
        (element->_ref_cnt)++;
        *refer_pntr = &(element->_ref_cnt);
        return(element->_str);
    }

    if (alloc_on_insert) {
        char *tmp = str;
        int   len = strlen(tmp) + 1;
        str = new char[len];
        memcpy(str,tmp,len);
    }

    element = new Cor_StrRefElement(str,1);
    *refer_pntr = &(element->_ref_cnt);

    element->set_next(_reftab[indx]);
    _reftab[indx] = element;

    if (++_num_elements > _element_limit)
        rehash_bigger();

    return(str);
}

// *************************************************************************
void Cor_StrRefTable::destroy(char *str, HashVal *hash)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!this)
        return;

    Cor_StrRefElement *before, *now;
    HashVal hval;

    if (hash)
        hval = *hash;
    else
        hval = Cor_StrHash(str);

    int indx = hval % _size;

    for (before = NULL, now = _reftab[indx]; now;
            before = now, now = now->get_next() ) {
    
        if ((hval == now->_hash) && (*str == *(now->_str))
                && (strcmp(str,now->_str) == 0)) {

              // Can be < 0 if some counting was done manually
              // (using the refer_pntr)
            if (--(now->_ref_cnt) <= 0) {

                if (before)
                    before->set_next(now->get_next());
                else
                    _reftab[indx] = now->get_next();

                delete now;
                _num_elements--;

            }
            break;
        }
    }
}

// *************************************************************************
// Returns NULL if string is not in table, otherwise the element its in.
// For items that hash to VarHashValue, returns NULL and sets the
// index to -1
Cor_StrRefElement *Cor_StrRefTable::lookup(char *str, int *my_indx,
        HashVal *hash, int hash_valid)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(this != NULL);

    Cor_StrRefElement *now;
    int indx;
    HashVal hval;

    if (hash_valid) {
        hval = *hash;
    } else {
        hval = Cor_StrHash(str);
        if (hash)
            *hash = hval;
    }

    if (!my_indx)
        my_indx = &indx;

    *my_indx = hval % _size;

    for (now = _reftab[*my_indx]; now; now = now->get_next() ) {

        if ((hval == now->_hash) && (*str == *(now->_str))
                && (strcmp(str,now->_str) == 0))
            return(now);
    }

    return(NULL);
}

// *************************************************************************
// Delete all items in the table
void Cor_StrRefTable::clear(int _complain)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(this != NULL);

    int j;
    int cmpln = _complain;
    Cor_StrRefElement *element, *prev = NULL;

    for (j=0; j < _size; j++) {
        
        for (element = _reftab[j]; element; element=element->get_next()) {
            delete prev;
            prev = element;
            if (element->_ref_cnt) {
                // Better to leak these at the moment
                if (cmpln)
                    cerr << "Leaking element for string " << element->_str
                         << endl;
                prev = NULL;
            }
        }

        delete prev;
        prev = NULL;
    }

    cmpln = 0;

        // Clear out the table in case its being re-used
    for (j=0; j < _size; j++)
        _reftab[j] = NULL;
}

// *************************************************************************
// Double the size of the hash table and rehash all elements
void Cor_StrRefTable::rehash_bigger()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(this != NULL);

    HashVal hval;
    int j, indx;

    int old_size = _size;
    Cor_StrRefElement **old_reftab = _reftab;

    _size = _size + _size;
    _element_limit = _size + _size;
    _reftab = new Cor_StrRefElement*[_size];

    for (j=0; j < _size; j++)
        _reftab[j] = NULL;


        // Rehash the old items into the new table
    for (j=0; j < old_size; j++) {

        Cor_StrRefElement *cur, *prev = NULL;
        for (cur = old_reftab[j]; cur; cur = cur->_next) {
     
            if (prev) {
     
                hval = prev->_hash;
                indx = hval % _size;

                prev->set_next(_reftab[indx]);
                _reftab[indx] = prev;
            }

            prev = cur;
        }

        if (prev) {
     
            hval = prev->_hash;
            indx = hval % _size;

            prev->set_next(_reftab[indx]);
            _reftab[indx] = prev;
        }
    }

    delete [] old_reftab;
}

// *************************************************************************
// For scanning the entire table
Cor_StrRefCursor *Cor_StrRefTable::OpenScan()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_StrRefCursor *cursor = new Cor_StrRefCursor(this);

    return(cursor);
}

// *************************************************************************
